@@ -164,20 +164,30 @@ make_menubar (void)
 	return menubar;
 }
 
+static void
+get_gdk_color_from_rgb (GdkColor *c, guint32 rgb)
+{
+	c->red = ((rgb & 0xff0000) >> 16) * 0x101;
+	c->green = ((rgb & 0xff00) >> 8) * 0x101;
+	c->blue = (rgb & 0xff) * 0x101;
+	c->pixel = 0;
+}
+
 void
 greeter_item_create_canvas_item (GreeterItemInfo *item)
 {
   GnomeCanvasGroup *group;
+  GtkJustification just;
   GtkWidget *entry;
   GtkWidget *list;
   GtkWidget *swin;
   double x1, y1, x2, y2;
   int i;
   GtkAllocation rect;
-  GtkJustification just;
   char *text;
   GtkTooltips *tooltips;
   char *num_locale;
+  GdkColor c;
 
   if (item->item != NULL)
     return;
@@ -217,7 +227,7 @@ greeter_item_create_canvas_item (GreeterItemInfo *item)
 					"y1", y1,
 					"x2", x2,
 					"y2", y2,
-					"fill_color_rgba", item->colors[GREETER_ITEM_STATE_NORMAL],
+					"fill_color_rgba", item->data.rect.colors[GREETER_ITEM_STATE_NORMAL],
 					NULL);
     break;
   case GREETER_ITEM_TYPE_SVG:
@@ -225,21 +235,21 @@ greeter_item_create_canvas_item (GreeterItemInfo *item)
     setlocale (LC_NUMERIC, "C");
     for (i = 0; i < GREETER_ITEM_STATE_MAX; i++)
       {
-	if (item->files[i] != NULL)
+	if (item->data.pixmap.files[i] != NULL)
 	  {
 	    if (i > 0 &&
-		item->files[0] != NULL &&
-		item->pixbufs[0] != NULL &&
-		strcmp (item->files[0], item->files[i]) == 0)
-	      item->pixbufs[i] = g_object_ref (item->pixbufs[0]);
+		item->data.pixmap.files[0] != NULL &&
+		item->data.pixmap.pixbufs[0] != NULL &&
+		strcmp (item->data.pixmap.files[0], item->data.pixmap.files[i]) == 0)
+	      item->data.pixmap.pixbufs[i] = g_object_ref (item->data.pixmap.pixbufs[0]);
 	    else
-	      item->pixbufs[i] =
-	        rsvg_pixbuf_from_file_at_size (item->files[i],
+	      item->data.pixmap.pixbufs[i] =
+	        rsvg_pixbuf_from_file_at_size (item->data.pixmap.files[i],
 					       rect.width, rect.height,
 					       NULL);
 	  }
 	else
-	  item->pixbufs[i] = NULL;
+	  item->data.pixmap.pixbufs[i] = NULL;
       }
     setlocale (LC_NUMERIC, num_locale);
     g_free (num_locale);
@@ -249,27 +259,27 @@ greeter_item_create_canvas_item (GreeterItemInfo *item)
   case GREETER_ITEM_TYPE_PIXMAP:
     for (i = 0; i < GREETER_ITEM_STATE_MAX; i++)
       {
-	GdkPixbuf *pb = item->pixbufs[i];
+	GdkPixbuf *pb = item->data.pixmap.pixbufs[i];
 	if (pb != NULL)
 	  {
-	    item->pixbufs[i] =
+	    item->data.pixmap.pixbufs[i] =
 	      transform_pixbuf (pb,
-				(item->have_tint & (1<<i)), item->tints[i],
-				item->alphas[i], rect.width, rect.height);
+				(item->data.pixmap.have_tint & (1<<i)), item->data.pixmap.tints[i],
+				(double)item->data.pixmap.alphas[i] / 256.0, rect.width, rect.height);
 	    g_object_unref (pb);
 	  }
       }
     
-    if (item->pixbufs[GREETER_ITEM_STATE_NORMAL] != NULL)
+    if (item->data.pixmap.pixbufs[GREETER_ITEM_STATE_NORMAL] != NULL)
       item->item = gnome_canvas_item_new (group,
 					  GNOME_TYPE_CANVAS_PIXBUF,
 					  "x", (gdouble) x1,
 					  "y", (gdouble) y1,
-					  "pixbuf", item->pixbufs[GREETER_ITEM_STATE_NORMAL],
+					  "pixbuf", item->data.pixmap.pixbufs[GREETER_ITEM_STATE_NORMAL],
 					  NULL);
     break;
   case GREETER_ITEM_TYPE_LABEL:
-    text = greeter_item_expand_text (item->orig_text);
+    text = greeter_item_expand_text (item->data.text.orig_text);
 
     /* Justification is taken from the anchor */
     if (item->anchor == GTK_ANCHOR_NORTH_WEST ||
@@ -285,20 +295,30 @@ greeter_item_create_canvas_item (GreeterItemInfo *item)
 
     item->item = gnome_canvas_item_new (group,
 					GNOME_TYPE_CANVAS_TEXT,
-					"markup", text,
+					"text", "",
 					"x", x1,
 					"y", y1,
 					"anchor", item->anchor,
-					"font_desc", item->fonts[GREETER_ITEM_STATE_NORMAL],
-					"fill_color_rgba", item->colors[GREETER_ITEM_STATE_NORMAL],
+					"font_desc", item->data.text.fonts[GREETER_ITEM_STATE_NORMAL],
+					"fill_color_rgba", item->data.text.colors[GREETER_ITEM_STATE_NORMAL],
 					"justification", just,
 					NULL);
 
+    greeter_canvas_item_break_set_string (item,
+					  text,
+					  TRUE /* markup */,
+					  item->data.text.real_max_width,
+					  NULL /* width */,
+					  NULL /* height */,
+					  NULL /* canvas */,
+					  item->item);
+    g_free (text);
+
     /* if there is an accelerator we do an INCREDIBLE hack */
-    if (strchr (item->orig_text, '_') != NULL)
+    if (strchr (item->data.text.orig_text, '_') != NULL)
       {
 	GreeterItemInfo *button;
-	GtkWidget *fake_button = gtk_button_new_with_mnemonic (item->orig_text);
+	GtkWidget *fake_button = gtk_button_new_with_mnemonic (item->data.text.orig_text);
 	gtk_widget_show (fake_button);
 	GTK_WIDGET_UNSET_FLAGS (fake_button, GTK_CAN_FOCUS);
 	gnome_canvas_item_new (gnome_canvas_root (GNOME_CANVAS (canvas)),
@@ -318,8 +338,6 @@ greeter_item_create_canvas_item (GreeterItemInfo *item)
 			       (GClosureNotify)g_free,
 			       0 /* connect_flags */);
       }
-
-    g_free (text);
     
     break;
     
@@ -328,6 +346,10 @@ greeter_item_create_canvas_item (GreeterItemInfo *item)
     gtk_entry_set_has_frame (GTK_ENTRY (entry), FALSE);
     if (GdmUseCirclesInEntry)
       gtk_entry_set_invisible_char (GTK_ENTRY (entry), 0x25cf);
+    gtk_widget_modify_font (entry, item->data.text.fonts[GREETER_ITEM_STATE_NORMAL]);
+
+    get_gdk_color_from_rgb (&c, item->data.text.colors[GREETER_ITEM_STATE_NORMAL]);
+    gtk_widget_modify_text (entry, GTK_STATE_NORMAL, &c);
     
     if (item->id != NULL && strcmp (item->id, "user-pw-entry") == 0) {
 	    /* HACK! Add a menubar, this is kind of evil isn't it,
@@ -399,3 +421,155 @@ greeter_item_create_canvas_item (GreeterItemInfo *item)
 		      item);
 }
 
+static gboolean
+append_word (GString *str, GString *line, GString *word, int max_width, const char *textattr, GnomeCanvasItem *canvas_item)
+{
+	int width, height;
+	char *try = g_strconcat (line->str, word->str, NULL);
+	gnome_canvas_item_set (GNOME_CANVAS_ITEM (canvas_item), textattr, try, NULL);
+	g_free (try);
+
+	pango_layout_get_pixel_size (GNOME_CANVAS_TEXT (canvas_item)->layout, &width, &height);
+
+
+	if (width > max_width) {
+		if ( ! ve_string_empty (line->str)) {
+			if (str->len > 0 &&
+			    str->str[str->len-1] == ' ') {
+				g_string_truncate (str, str->len-1);
+			}
+			g_string_append_unichar (str, '\n');
+		}
+		g_string_assign (line, word->str);
+		g_string_append (str, word->str);
+		g_string_assign (word, "");
+
+		return TRUE;
+	} else {
+		g_string_append (line, word->str);
+		g_string_append (str, word->str);
+		g_string_assign (word, "");
+
+		return FALSE;
+	}
+}
+
+void
+greeter_canvas_item_break_set_string (GreeterItemInfo *info,
+				      const char *orig,
+				      gboolean markup,
+				      int max_width,
+				      int *width,
+				      int *height,
+				      GnomeCanvas *canvas,
+				      GnomeCanvasItem *real_item)
+{
+	PangoLogAttr *attrs;
+	int n_chars;
+	GString *str;
+	GString *word;
+	GString *line;
+	int i;
+	int n_attrs;
+	int ia;
+	const char *p;
+	int in_current_row;
+	GnomeCanvasItem *canvas_item;
+	const char *textattr = markup ? "markup" : "text";
+
+	str = g_string_new (NULL);
+	word = g_string_new (NULL);
+	line = g_string_new (NULL);
+
+	/* A gross hack */
+	if (real_item != NULL)
+		canvas_item = real_item;
+	else
+		canvas_item = gnome_canvas_item_new (gnome_canvas_root (canvas),
+						     GNOME_TYPE_CANVAS_TEXT,
+						     textattr, "",
+						     "x", 0.0,
+						     "y", 0.0,
+						     "font_desc", info->data.text.fonts[GREETER_ITEM_STATE_NORMAL],
+						     NULL);
+
+	if (max_width == 0) {
+		gnome_canvas_item_set (GNOME_CANVAS_ITEM (canvas_item), textattr, orig, NULL);
+		pango_layout_get_pixel_size (GNOME_CANVAS_TEXT (canvas_item)->layout, width, height);
+
+		if (real_item != canvas_item)
+			gtk_object_destroy (GTK_OBJECT (canvas_item));
+		return;
+	}
+
+	n_chars = g_utf8_strlen (orig, -1);
+
+	gnome_canvas_item_set (GNOME_CANVAS_ITEM (canvas_item), textattr, orig, NULL);
+	pango_layout_get_log_attrs (GNOME_CANVAS_TEXT (canvas_item)->layout, &attrs, &n_attrs);
+
+	i = 0;
+	ia = 0;
+	in_current_row = 0;
+	p = orig;
+	while (i < n_chars) {
+		gunichar ch;
+
+		ch = g_utf8_get_char (p);
+
+		if (markup && ch == '<') {
+			while (i < n_chars) {
+				g_string_append_unichar (word, ch);
+				p = g_utf8_next_char (p);
+				i ++;
+
+				if (ch == '>') {
+					ch = g_utf8_get_char (p);
+					break;
+				} else {
+					ch = g_utf8_get_char (p);
+				}
+			}
+			if (i >= n_chars)
+				break;
+		}	
+
+		if (attrs[ia].is_line_break && in_current_row > 0) {
+			if (append_word (str, line, word, max_width, textattr, canvas_item))
+				in_current_row = 0;
+		}
+
+		in_current_row ++;
+		g_string_append_unichar (word, ch);
+
+		p = g_utf8_next_char (p);
+		i ++;
+		ia ++;
+
+		/* eeek! */
+		if (ia >= n_attrs) {
+			while (i < n_chars) {
+				ch = g_utf8_get_char (p);
+				g_string_append_unichar (word, ch);
+				p = g_utf8_next_char (p);
+				i ++;
+			}
+			break;
+		}
+	}
+
+	if ( ! ve_string_empty (word->str))
+		append_word (str, line, word, max_width, textattr, canvas_item);
+
+	gnome_canvas_item_set (GNOME_CANVAS_ITEM (canvas_item), textattr, str->str, NULL);
+	pango_layout_get_pixel_size (GNOME_CANVAS_TEXT (canvas_item)->layout, width, height);
+
+	if (real_item != canvas_item)
+		gtk_object_destroy (GTK_OBJECT (canvas_item));
+	g_free (attrs);
+
+	g_string_free (line, TRUE);
+	g_string_free (word, TRUE);
+
+	g_string_free (str, TRUE);
+}
+
